/******************************************************

    Menu and viewer for ASCII user interface.
    High score table too for some reason.

******************************************************/

#ifndef _PANEL_H
extern "C" {
#include <panel.h>
}
#define _PANEL_H
#endif

#include <stdio.h>
#include <ctype.h>
#include "Util.h"
#include "AUI.h"
#include "Game.h"
#include "Menu.h"


shASCIIMenu::shASCIIMenu (const char *prompt, int flags)
    : mChoices ()
{
    strncpy (mPrompt, prompt, 79); mPrompt[79] = 0;
    mFlags = flags;
    mResultIterator = 0;
    mHeight = 0;
    mOffset = 0;
    mDone = 0;
    getmaxyx(stdscr, mHeight, mWidth);

    if (!(mFlags & kNoPick)) {
        /* KLUDGE: avoid repeating a letter on the same page */
        mHeight = mini (52, mHeight);
    }
    mObjTypeHack = kMaxObjectType;
    mHelpFileName = NULL;
}


shASCIIMenu::~shASCIIMenu ()
{
    for (int i = 0; i < mChoices.count (); ++i) {
        delete mChoices.get (i);
    }
}


void
shASCIIMenu::addIntItem (char letter, const char *text, int value,
                    int count /* = 1 */, int selected /* = 0 */)
{
    shMenuChoice *c;

    if (0 == letter) {
        letter = ' ';
    }

    c = new shMenuChoice (letter, text, value, count, selected);
    mChoices.add (c);
}


void
shASCIIMenu::addPtrItem (char letter, const char *text, const void *value,
                    int count /* = 1 */, int selected /* = 0 */)
{
    const char *typeHeader[kMaxObjectType] =
    {
        "UNINITIALIZED",
        "Money               (toggle all with $)",
        "Bionic Implants     (toggle all with :)",
        "Floppy Disks        (toggle all with ?)",
        "Canisters           (toggle all with !)",
        "Tools               (toggle all with ()",
        "Armor               (toggle all with [)",
        "Weapons             (toggle all with ))",
        "Ammunition          (toggle all with =)",
        "Devices             (toggle all with &)",
        "Ray Guns            (toggle all with /)",
        "Energy Cells        (toggle all with *)"
    };

    shMenuChoice *c;

    if (0 == letter) {
        letter = ' ';
    }
    if (mFlags & kCategorizeObjects && value) {
        shObjectType t = ((shObject *) value) -> myIlk ()->mType;
        if (t != mObjTypeHack) {
            if (t >= kMoney && t <= kEnergyCell) {
                if (mFlags & kMultiPick) { /* Show toggle key. */
                    addHeader (typeHeader[t]);       /* From above. */
                } else {
                    addHeader (objectTypeHeader[t]); /* Default. */
                }
            } else {
                addHeader ("------");
            }
            mObjTypeHack = t;
        }
    }

    c = new shMenuChoice (letter, text, value, count, selected);
    mChoices.add (c);
}


void
shASCIIMenu::addHeader (const char *text)
{   /* prints out the header */
    addPtrItem (-1, text, NULL, -1);
}

void
shASCIIMenu::addPageBreak ()
{   /* Just a prettifier. */
    addPtrItem (0, "", NULL, -1);
    /* -2 is signal to adjust space on both sides of text. */
    addPtrItem (-2, "---..oo..--=oOo=--..oo..---", NULL, -1);
    addPtrItem (0, "", NULL, -1);
}

void
shASCIIMenu::finish () {
    assert (kNoPick & mFlags);
    accumulateResults ();
}


shMenuChoice *
shASCIIMenu::getResultChoice ()
{
    if (!mDone)
        accumulateResults ();

    while (mResultIterator < mChoices.count ()) {
        shMenuChoice *choice = mChoices.get (mResultIterator++);
        if ((choice->mSelected)) {
            return choice;
        }
    }
    return NULL;
}

/* call this repeatedly to store the selected results into value and count.
  RETURNS: 0 if there are no (more) results, 1 o/w
*/

int
shASCIIMenu::getIntResult (int *value, int *count /* = NULL */ )
{
    shMenuChoice *choice = getResultChoice ();

    if (choice) {
        *value = choice->mValue.mInt;
        if (NULL != count) {
            *count = choice->mSelected;
        }
        return 1;
    }
    *value = 0;
    return 0;
}


int
shASCIIMenu::getPtrResult (const void **value, int *count /* = NULL */ )
{
    shMenuChoice *choice = getResultChoice ();
    if (mDone == DELETE_FILTER_SIGNAL)
        return DELETE_FILTER_SIGNAL;

    if (choice) {
        *value = choice->mValue.mPtr;
        if (NULL != count) {
            *count = choice->mSelected;
        }
        return 1;
    }
    *value = NULL;
    return 0;
}


void
shASCIIMenu::attachHelp (const char *fname)
{
    mHelpFileName = fname;
}


void
shASCIIMenu::select (int i1, int i2,   /* mChoices[i1..i2) */
                int action,       /* 0 unselect, 1 select, 2 toggle */
                shObjectType t)   /* = kUninitialized */
{
    int i;
    if (!(mFlags & kMultiPick)) {
        return;
    }

    for (i = i1; i < i2; i++) {
        shMenuChoice *choice = mChoices.get (i);
        if (choice->mCount < 0) {
            continue;
        }
        if (t &&
            mFlags & kCategorizeObjects &&
            choice->mValue.mPtr &&
            t != ((shObject *) choice->mValue.mPtr) -> myIlk ()->mType)
        {
            continue;
        }
        if (0 == action) {
            choice->mSelected = 0;
        } else if (1 == action) {
            choice->mSelected = choice->mCount;
        } else if (2 == action) {
            if (!choice->mSelected)
                choice->mSelected = choice->mCount;
            else
                choice->mSelected = 0;
        }
    }
}


const char **
shASCIIMenu::prepareHelp (int *lines)
{   /* Building blocks for help. */
    const char *navigation =
"ARROWS, PAGE UP, PAGE DOWN navigate";
    const char *navigation_more =
"ARROWS, PAGE UP, PAGE DOWN navigate   BACKSPACE disable filter";
    const char *navigation_count =
"ARROWS, PAGE UP, PAGE DOWN navigate   0..9 specify count";
    const char *nopick =
        "SPACE, ENTER, ESCAPE  close window";
    const char *singlepick =
"SPACE, ENTER, ESCAPE  cancel  LETTER  choose and accept";
    const char *multipick =
"Items:  , select all  - deselect all  @ toggle all  LETTER toggle single";
    const char *multipick_nocount_last =
"SPACE, ENTER, ESCAPE accept";
    const char *multipick_count_last =
"SPACE, ENTER, ESCAPE accept  0..9 specify count";
    const char *help_available =
        "TAB  show help file";
    /* Choose appropriate blocks. */
    *lines = 0;
    const char **text = (const char **) calloc (6, sizeof (char *));
    if (mFlags & kNoPick) {
        *lines = 2;
        text[0] = navigation;
        text[1] = nopick;
    } else if (mFlags & kMultiPick) {
        *lines = 3;
        text[0] = navigation;
        text[1] = multipick;
        if (mFlags & kCountAllowed) {
            text[2] = multipick_count_last;
        } else {
            text[2] = multipick_nocount_last;
        }
    } else if (!(mFlags & kNoHelp)) {
        *lines = 2;
        if (mFlags & kCountAllowed) {
            text[0] = navigation_count;
        } else if (mFlags & kAnythingAllowed) {
            text[0] = navigation_more;
        } else {
            text[0] = navigation;
        }
        text[1] = singlepick;
    } else {
        *lines = 0;
    }
    if (mHelpFileName) {
        text[*lines] = help_available;
        ++(*lines);
    }
    return text;
}


void
shASCIIMenu::showHelp (WINDOW *win)
{
    int lines;
    const char **text = prepareHelp (&lines);

    for (int i = 0; i < lines; ++i) {
        int len = strlen (text[i]);
        for (int j = 0; j < len; ++j) {
            if (islower (text[i][j]) || text[i][j] == ':' ||
                (j == 0 && (i == 1 && (mFlags & kMultiPick))))
            {
                wattrset (win, A_NORMAL);
            } else {
                wattrset (win, A_BOLD);
            }
            mvwaddnstr (win, i, j, &text[i][j], 1);
        }
    }
    free (text);
}


void
shASCIIMenu::dropResults ()
{
    mDone = 0;
    mResultIterator = 0;
    for (int i = 0; i < mChoices.count (); ++i) {
        mChoices.get (i) -> mSelected = 0;
    }
}

/* Returns true when key pressed was valid. */
bool
shASCIIMenu::interpretKey (int key, shInterface::Command cmd)
{
    if (cmd) switch (cmd) {
    case shInterface::kDrop: /* Drop item filter. */
        if (kAnythingAllowed & mFlags) {
            mDone = DELETE_FILTER_SIGNAL;
            return true;
        }
        return false;
    case shInterface::kMoveNW: /* Home */
        mLast -= mOffset;
        mOffset = 0;
        return true;
    case shInterface::kMoveSW: /* End */
        mOffset += mChoices.count () - mLast;
        mLast = mChoices.count ();
        return true;
    case shInterface::kMoveUp:
        if (mOffset > 0) {
            --mOffset;
            --mLast;
        }
        return true;
    case shInterface::kMoveDown:
        if (mLast < mChoices.count ()) {
            ++mOffset;
            ++mLast;
        }
        return true;
    case shInterface::kMoveNE: /* Page up */
        mOffset -= mItemHeight / 2;
        mLast -= mItemHeight / 2;
        if (mOffset < 0) {
            mLast -= mOffset;
            mOffset = 0;
        }
        return true;
    case shInterface::kMoveSE: /* Page down */
        mLast += mItemHeight / 2;
        mOffset += mItemHeight / 2;
        if (mLast >= mChoices.count ()) {
            mOffset -= mLast - mChoices.count ();
            mLast = mChoices.count ();
        }
        return true;
    case shInterface::kHelp:
        {
            shTextViewer *viewer = I->newViewer (mHelpFileName);
            viewer->show ();
            delete viewer;
        }
        return true;
    default: break;
    }
    switch (key) {
    case '@': /* toggle all */
        select (0, mChoices.count(), 2);
        return true;
    case '-': /* deselect all */
        select (0, mChoices.count(), 0);
        return true;
    case ',': /* select all */
        select (0, mChoices.count(), 1);
        return true;
    case '$': /* toggle all cash */
        select (0, mChoices.count(), 2, kMoney);
        return true;
    case ':': /* toggle all implants */
        select (0, mChoices.count(), 2, kImplant);
        return true;
    case '?': /* toggle all floppies */
        select (0, mChoices.count(), 2, kFloppyDisk);
        return true;
    case '!': /* toggle all cans */
        select (0, mChoices.count(), 2, kCanister);
        return true;
    case '(': /* toggle all tools */
        select (0, mChoices.count(), 2, kTool);
        return true;
    case '[': /* toggle all armor */
        select (0, mChoices.count(), 2, kArmor);
        return true;
    case ')': /* toggle all weapons */
        select (0, mChoices.count(), 2, kWeapon);
        return true;
    case '=': /* toggle all ammo */
        select (0, mChoices.count(), 2, kProjectile);
        return true;
    case '/': /* toggle all ray guns */
        select (0, mChoices.count(), 2, kRayGun);
        return true;
    case '*': /* toggle all energy */
        select (0, mChoices.count(), 2, kEnergyCell);
        return true;
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
        if (mFlags & kCountAllowed) {
            mNum = mNum * 10 + key - '0';
            if (mNum >= 1000000) {
                mNum = 1000000;
            }
            return true;
        } else { /* TODO: tell user count not allowed for this command. */
            return false;
        }
    default:
        for (int i = 0; i < mChoices.count (); ++i) {
            shMenuChoice *item = mChoices.get (i);
            if (item->mLetter == key) {
                if ((mFlags & kCountAllowed) && mNum) { /* Set exact number. */
                    if (mNum > item->mCount) {
                        /* TODO: Warn user not that many items are available. */
                        mNum = item->mCount;
                    }
                } else if (mFlags & kSelectIsPlusOne) {
                    mNum = item->mSelected == item->mCount ? 0 :
                           item->mSelected + 1;
                } else { /* Toggle all or none. */
                    mNum = item->mSelected ? 0 : item->mCount;
                }
                item->mSelected = mNum;
                mNum = 0;
                if (!(mFlags & kMultiPick)) {
                    mDone = 1;
                }
                return true;
            }
        }
        return false;
    }
    return false;
}


void
shASCIIMenu::accumulateResults ()
{
    const int map_end = 64; /* column map ends */
    WINDOW *win, *helpwin = NULL;
    PANEL *panel, *helppanel = NULL;

    int helplines;
    if (!(mFlags & kNoHelp))
        if (mFlags & kMultiPick)    helplines = 3;
        else                        helplines = 2;
    else                            helplines = 0;
    if (mHelpFileName) ++helplines;

    /* Two additional lines besides all the choices are the menu title
       and --End-- or --More-- at the bottom. */
    mItemHeight = mini (mHeight - helplines, mChoices.count () + 2);
    int width = 10;
    int gap = 0; /* Used to position the menu. */
    for (int i = 0; i < mChoices.count (); ++i) {
        width = maxi (width, strlen (mChoices.get (i)->mText) + 1);
    }
    if (!(mFlags & kNoPick)) width += 10; /* Adjust for "( ) x - " prompts. */
    /* Main header might be still longer. */
    width = maxi (width, strlen (mPrompt) + 2);
    if (mFlags & kCategorizeObjects) { /* Do not center item lists. */
        /* Determine whether window can leave sidebar unobscured. */
        if (width <= map_end) { /* Yes! */
            width = map_end;
        } else {                /* No, so hide it whole. */
            width = mWidth;
        }
    } else {
        width = mini (mWidth, width);
        if (width > map_end) { /* Would obscure side bar window? */
            width = mWidth;    /* Then cover it whole. */
        } else { /* Place small gap between sidebar and menu if possible. */
            gap = mini (10, (map_end - width) / 2);
        }
    }

    win = newwin (mItemHeight, width, 0, maxi (0, map_end - width) - gap);
    if (!win) {
        debug.log ("Unable to create window (%d, %d, %d, %d)",
                  mItemHeight, width, 0, maxi (0, map_end - width));
        I->p ("Uh oh!  Couldn't create window!!");
        mDone = 1;
        return;
    }
    keypad (win, TRUE);
    panel = new_panel (win);
    if (helplines) {
        helpwin = newwin (helplines, 80, mHeight - helplines, 0);
        if (!helpwin) {
            debug.log ("Unable to create help window (%d, %d, %d, %d)",
                      helplines, 80, mHeight - helplines, 0);
            I->p ("Uh oh!  Couldn't create help window!!");
            mDone = 1;
            return;
        }
        helppanel = new_panel (helpwin);
        showHelp (helpwin);
    }

    /* -2 lines to make space for header and --End-- or similar. */
    mLast = mini (mOffset + mItemHeight - 2, mChoices.count ());
    while (1) { /* Menu loop. */
        /* Menu header: */
        wattrset (win, A_BOLD);
        mvwaddnstr (win, 0, 1, mPrompt, width);
        wclrtoeol (win);

        wattrset (win, A_NORMAL);
        int i;
        for (i = mOffset; i < mLast; ++i) {
            /* First, clear line. */
            wmove (win, 1 + i - mOffset, 0);
            wclrtoeol (win);
            /* Then draw. */
            char buf[100];
            shMenuChoice *item = mChoices.get (i);

            if (item->mLetter >= 0 && (shMenu::kCategorizeObjects & mFlags)
                && ((shObject *) item->mValue.mPtr)->isKnownRadioactive ())
            {
                wattrset (win, ColorMap[kGreen]);
            }
            if (-2 == item->mLetter) { /* This is a pretty delimiter. */
                int len = strlen (item->mText);
                int j = (width - len) / 2;
                char *spaces = GetBuf ();
                for (int i = 0; i < j; ++i) {
                    spaces[i] = ' ';
                }
                spaces[j] = '\0';
                snprintf (buf, 100, "%s%s%s", spaces, item->mText, spaces);
            } else if (-1 == item->mLetter) { /* This is a header entry. */
                if (kCategorizeObjects & mFlags && mFlags & kMultiPick) {
                    /* Get category header. */
                    char part[50];
                    snprintf (part, 50, " %s ", item->mText);
                    char *gap = strstr (part, "  "); /* Find gap. */
                    gap[1] = 0; /* Truncate. */
                    wattrset (win, A_REVERSE);
                    mvwaddnstr (win, 1 + i - mOffset, 1, part, width);
                    /* Get (toggle all with X) part. */
                    char *p2 = strstr (item->mText, "(t");
                    int len = strlen (p2);
                    snprintf (part, len-2, "%s", item->mText);
                    wattrset (win, ColorMap[kBlue]);
                    mvwaddnstr (win, 1 + i - mOffset, width-21, p2, width);
                    /* The toggle key should stand out. */
                    snprintf (buf, len - 1, "%c", p2[len - 2]);
                    mvwaddch (win, 1 + i - mOffset, width-21+len-2,
                              p2[len - 2] | ColorMap[kWhite]);
                    buf[0] = 0; /* Printing is done. */
                } else {
                    wattrset (win, A_REVERSE);
                    snprintf (buf, 100, " %s ", item->mText);
                }
            } else if (mFlags & kNoPick) {
                if (' ' == item->mLetter) {
                    snprintf (buf, 100, "%s", item->mText);
                } else {
                    snprintf (buf, 100, "%c - %s", item->mLetter, item->mText);
                }
            } else {
                if (' ' == item->mLetter) {
                    snprintf (buf, 100, "        %s", item->mText);
                } else if (mFlags & kShowCount) {
                    if (item->mSelected) {
                        snprintf (buf, 100, "(%d) %c - %s",
                            item->mSelected, item->mLetter, item->mText);
                    } else {
                        snprintf (buf, 100, "( ) %c - %s",
                            item->mLetter, item->mText);
                    }
                } else {
                    snprintf (buf, 100, "(%c) %c - %s",
                              0 == item->mSelected ? ' ' :
                              item->mCount == item->mSelected ? 'X' : '#',
                              item->mLetter, item->mText);
                }
            }
            mvwaddnstr (win, 1 + i - mOffset, 1, buf, width);
            wattrset (win, A_NORMAL);
        }
        if (mLast == mChoices.count () && mOffset == 0) {
            mvwaddnstr (win, 1 + i - mOffset, 1, "--All shown--", width);
        } else if (i == mChoices.count ()) {
            mvwaddnstr (win, 1 + i - mOffset, 1, "--End--", width);
        } else if (0 == mOffset) {
            mvwaddnstr (win, 1 + i - mOffset, 1, "--Top--", width);
        } else {
            mvwaddnstr (win, 1 + i - mOffset, 1, "--More--", width);
        }
        update_panels ();
        doupdate();

        while (1) {
            //update_panels ();
            //doupdate();
            int key = I->getChar ();

            if (27 == key || 13 == key || ' ' == key) { /* done */
                mDone = 1; break;
            } else if (KEY_BACKSPACE == key) {
                interpretKey (0, shInterface::kDrop); break;
            } else if (KEY_HOME == key || KEY_A1 == key) {
                interpretKey (0, shInterface::kMoveNW); break;
            } else if (KEY_END == key || KEY_C1 == key) {
                interpretKey (0, shInterface::kMoveSW); break;
            } else if (KEY_UP == key) {
                interpretKey (0, shInterface::kMoveUp); break;
            } else if (KEY_DOWN == key) {
                interpretKey (0, shInterface::kMoveDown); break;
            } else if (KEY_PPAGE == key || KEY_LEFT == key || KEY_A3 == key) {
                interpretKey (0, shInterface::kMoveNE); break;
            } else if (KEY_NPAGE == key || KEY_RIGHT == key || KEY_C3 == key) {
                interpretKey (0, shInterface::kMoveSE); break;
            } else if ('\t' == key && mHelpFileName) { /* invoke help */
                interpretKey (0, shInterface::kHelp); break;
            } else if (mFlags & kNoPick) {
                continue;
            } else {
                if (interpretKey (key)) break;
            }
            if (mDone) break;
        }
        if (mDone) break;
    }
    /* Clean up. */
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    if (helplines) {
        hide_panel (helppanel);
        del_panel (helppanel);
        delwin (helpwin);
    }
    update_panels ();
    I->drawScreen ();
}



shASCIITextViewer::shASCIITextViewer (const char *fname)
{   /* Try opening target file. */
    assert (fname);
    mLines = NULL;
    mNumLines = 0;
    mMaxWidth = 0;
    char path[PRIME_PATH_LENGTH];
    snprintf (path, sizeof (path)-1, "%s/%s", DATADIR, fname);
    FILE *text = fopen (path, "r");
    if (!text) return;
    /* Count lines and check max width. */
    char *buf = GetBuf ();
    while (fgets (buf, SHBUFLEN, text)) {
        mMaxWidth = maxi (mMaxWidth, strlen (buf));
        ++mNumLines;
    }
    ++mNumLines; /* For adding " --End-- " at the bottom. */
    mMaxWidth = maxi (mMaxWidth, 9 /* --End-- */); /* Ok, this is paranoidal. */
    ++mMaxWidth; /* Compensate for \0. */
    /* Copy file to memory. */
    mLines = (char *) calloc (mNumLines * mMaxWidth, sizeof (char));
    rewind (text);
    int line = 0;
    while (fgets (mLines + line * mMaxWidth, mMaxWidth, text)) {
        ++line;
    }
    strcpy (mLines + line * mMaxWidth, " --End-- ");
    /* Done! Now clean up. */
    fclose (text);
}

shASCIITextViewer::shASCIITextViewer (char *lines, int num, int max)
{   /* Import text from some other procedure. */
    mLines = lines;
    mNumLines = num;
    mMaxWidth = max;
}

     /* Prints a single line of file to screen. Handles special */
void /* color change sequences: @X where X stands for a color code. */
shASCIITextViewer::print (WINDOW *win, int winline, int fileline)
{
    int control = 0, idx = 0;
    char *line = mLines + fileline * mMaxWidth;
    wmove (win, winline, 0);
    while (line[idx]) { /* Parse line. */
        if (control) {
            wattrset (win, ColorMap[line[idx] - 'A']);
            control = 0;
        } else if (line[idx] == '@') {
            control = 1;
        } else {
            waddch (win, line[idx]);
        }
        ++idx;
    }
}
     /* Displays loaded file to the screen. This procedure */
void /* should have no reason to modify member variables.  */
shASCIITextViewer::show (void)
{
    int height = LINES - 1;
    WINDOW *win;
    win = newwin (LINES, COLS, 0, 0); /* Take up full screen. */
    keypad (win, TRUE);
    if (!win) {
        debug.log ("Unable to create window (%d, %d, 0, 0)", LINES, COLS);
        I->p ("Uh oh!  Could not create window.");
        return;
    }
    PANEL *panel = new_panel (win);

    int first = 0;
    while (1) {
        /* Show content. */
        werase (win);
        int lines_drawn = mini (mNumLines - first, height);
        for (int i = first; i < first + lines_drawn; ++i) {
            print (win, i - first, i);
            //mvwaddstr (win, i - first, 0, mLines + i * mMaxWidth);
        }
        /* Show help at bottom. */
        wattrset (win, A_BOLD);
        mvwaddstr (win, height, 0,
"         SPACE  ENTER  ESCAPE          ARROWS  PAGE UP  PAGE DOWN");
        wattrset (win, A_NORMAL);
        mvwaddstr (win, height, 14, ",");
        mvwaddstr (win, height, 21, ",");
        mvwaddstr (win, height, 29, " - exit");
        mvwaddstr (win, height, 45, ",");
        mvwaddstr (win, height, 54, ",");
        mvwaddstr (win, height, 65, " - navigation");
        update_panels ();
        doupdate();
        int key = I->getChar ();
        /* Navigation. */
        if (KEY_HOME == key || KEY_A1 == key) {
            first = 0;
        } else if (KEY_END == key || KEY_C1 == key) {
            first = mNumLines - height;
        } else if (KEY_UP == key) {
            if (first > 0) --first;
        } else if (KEY_DOWN == key) {
            if (first + height < mNumLines) ++first;
        } else if (KEY_PPAGE == key || KEY_LEFT == key || KEY_A3 == key) {
            first -= height / 2;
            if (first < 0) first = 0;
        } else if (KEY_NPAGE == key || KEY_RIGHT == key || KEY_C3 == key) {
            first += height / 2;
            if (first + height > mNumLines) first = mNumLines - height;
        } else if (27 == key || 13 == key || ' ' == key) { /* Close. */
            break;
        }
    }
    /* Clean up. */
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    update_panels ();
    I->drawScreen ();
}

shASCIITextViewer::~shASCIITextViewer (void)
{
    free (mLines);
}



static int
copystr (char *dest, char *src, int len)
{
    int i;

    for (i = 0; i < len; i++) {
        dest[i] = src[i];
        if (0 == src[i]) {
            return -1;
        }
    }
    while (i > 0) {
        --i;
        if (isspace (src[i])) {
            dest[i] = 0;
            return i+1;
        }
    }
    dest[i] = 0;
    return len;
}

static void
displayScoreEntry (HighScoreEntry *e, int rank, int yours, WINDOW *win, int y)
{
    char buf[200];
    char buf2[78];
    int l, n;

    if (yours) {
        wattrset (win, ColorMap[kBrightRed]);
    } else {
        wattrset (win, ColorMap[kGray]);
    }
    if (rank >= 0) {
        snprintf (buf, 200, "%d.", rank + 1);
        mvwaddstr (win, y, 2, buf);
    }

    snprintf (buf, 200, "%8d", e->mScore);
    mvwaddstr (win, y, 6, buf);
    snprintf (buf, 200, "%s, %s", e->mName, e->mMessage);

    n = 0;
    do {
        l = copystr (buf2, &buf[n], 60);
        n += l;
        mvwaddstr (win, y, 15, buf2);
        y++;
    } while (l >= 0);

}


/* From Tombstone.cpp. */
void readHighScoreTable (HighScoreEntry scores[NUMHIGHSCORES],
                         int *nscores, int *entry);

void
shASCIIInterface::showHighScores ()
{
    HighScoreEntry scores[NUMHIGHSCORES+1];
    int nscores = 0;
    int entry = 0;

    readHighScoreTable (scores, &nscores, &entry);
    if (nscores == 0) {
        p ("The high score table is empty.");
        return;
    }

    /* Time to display high score. */
    WINDOW *win;
    PANEL *panel;
    int row = 1;

    win = newwin (25, 80, 0, 0);
    if (!win) {
        p ("Couldn't display high scores.");
        return;
    }
    panel = new_panel (win);

    if (GameOver && !GodMode) {
        if (NUMHIGHSCORES == entry) {
            mvwaddstr (win, row, 1, "You didn't make the high score list.");
            row += 2;
            displayScoreEntry (&scores[NUMHIGHSCORES], -1, 1, win, row);
        } else {
            mvwaddstr (win, row, 1, "You made the high score list!");
        }
    }
    if (GodMode) entry = NUMHIGHSCORES; /* Disable highlight. */
    for (int i = 0; i < nscores; ++i) {
        row+=2;
        displayScoreEntry (&scores[i], i, entry == i, win, row);
    }
    wattrset (win, ColorMap[kGray]);
    mvwaddstr (win, row + 2, 1, "--End--");

    update_panels ();
    doupdate();
    getChar ();
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    update_panels ();
    drawScreen ();
}


shMenu *
shASCIIInterface::newMenu (const char *prompt, int flags) {
    return new shASCIIMenu (prompt, flags);
}

shTextViewer *
shASCIIInterface::newViewer (const char *fname) {
    return new shASCIITextViewer (fname);
}

shTextViewer *
shASCIIInterface::newViewer (char *lines, int num, int max) {
    return new shASCIITextViewer (lines, num, max);
}
