#include "Global.h"
#include "Util.h"
#include "Object.h"
#include "Hero.h"
#include "Creature.h"
#include "ObjectType.h"
#include <math.h>

static int selectWeaponFireMode (shObject *gun);
static int useAnalProbe (shObject *probe);
static int chargePlasmaCaster (shObject *caster);

void
initializeWeapons ()
{
    /* enum: kObjAnyWeapon */
    /* enum: kObjFirstWeapon = kObjAnyWeapon */
    NEW_GENERIC_WEAPON(kObjAnyWeapon,
        "weapon", "weapon", "weapon",
        kCyan, kNoSkillCode, kNoSkillCode, kObjNothing, kSteel,
        0, 0, kTiny, 0, 0, +0,
        NULL, NULL, NULL, 0 BZ, ABSTRACT);

    /* enum: kObjGenericMissileWeapon */
    NEW_MISSILE_WEAPON(kObjGenericMissileWeapon,
        "missile weapon", "weapon", "weapon",
        kCyan, kObjAnyWeapon, kSteel,
        0, 0, kTiny, 0, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjShuriken */
    NEW_MISSILE_WEAPON(kObjShuriken,
        "shuriken", "throwing star", "pointy star",
        kCyan, kObjGenericMissileWeapon, kSteel,
        kMergeable, 100, kTiny, 10, 1, +0,
        newMissileAttack (shAttack::kStab, HALFTURN, 15, kPiercing, 1, 4),
        10 BZ, 10);

    /* enum: kObjSmartDisc */
    NEW_GENERIC_WEAPON(kObjSmartDisc,
        "Smart-Disc", "throwing disc", "throwing disc",
        kGreen, kMeleeWeapon, kNoSkillCode, kObjGenericMissileWeapon, kPlasteel,
        0, 650, kSmall, 30, 1, +0,
        newMeleeAttack (shAttack::kSlash, FULLTURN, kPiercing, 2, 4),
        newAttack (shAttack::kDisc, shAttack::kSmartDisc,
            kMissile, FULLTURN, 15, 0, kPiercing, 2, 7),
        NULL, 1000 BZ, 5);
    AllIlks[kObjSmartDisc]->mGlyph.mSym = '*';

    /* enum: kObjFootball */
    NEW_MISSILE_WEAPON(kObjFootball,
        "football", "spheroid", "prolate spheroid",
        kBrown, kObjGenericMissileWeapon, kLeather,
        kMergeable, 400, kSmall, 3, 1, +0,
        newMissileAttack (shAttack::kImpact, FULLTURN, 10, kConcussive, 1, 6),
        25 BZ, 5);

    /* enum: kObjGenericMeleeWeapon */
    NEW_MELEE_WEAPON(kObjGenericMeleeWeapon,
        "melee weapon", "weapon", "weapon",
        kCyan, kMeleeWeapon, kObjAnyWeapon, kSteel,
        0, 100, kTiny, 10, 1, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjElvenDagger */
    NEW_GENERIC_WEAPON(kObjElvenDagger,
        "elven dagger", "dagger", "wooden dagger",
        kCyan, kMeleeWeapon, kNoSkillCode, kObjGenericMeleeWeapon, kWood,
        kMergeable, 400, kTiny, 10, 1, +0,
        newMeleeAttack (shAttack::kStab, FULLTURN, kPiercing, 1, 4),
        newMissileAttack (shAttack::kStab, FULLTURN, 10, kPiercing, 1, 4),
        NULL, 10 BZ, 5);

    /* enum: kObjAnalProbe */
    NEW_MELEE_WEAPON(kObjAnalProbe,
        "anal probe", "anal probe", "anal probe",
        kGray, kMeleeWeapon, kObjGenericMeleeWeapon, kSteel,
        kPreidentified, 400, kTiny, 10, 1, +0,
        newMeleeAttack (shAttack::kAnalProbe, FULLTURN, kViolating, 10, 100),
        10 BZ, 0);
    AllIlks[kObjAnalProbe]->mUseFunc = useAnalProbe;

    /* enum: kObjClub */
    NEW_MELEE_WEAPON(kObjClub,
        "club", "club", "club",
        kBrown, kMeleeWeapon, kObjGenericMeleeWeapon, kWood,
        kPreidentified, 1200, kSmall, 5, 10, +0,
        newMeleeAttack (shAttack::kClub, FULLTURN, kConcussive, 1, 6),
        1 BZ, 5);

    /* enum: kObjCattleProd */
    NEW_MELEE_WEAPON(kObjCattleProd,
        "cattle prod", "energy baton", "cattle prod",
        kGray, kMeleeWeapon, kObjGenericMeleeWeapon, kSteel,
        0, 400, kSmall, 10, 2, +0,
        newMeleeAttackP (shAttack::kZap, FULLTURN,
            100, kElectrical, 1, 4,   50, kStunning, 1, 2),
        50 BZ, 5);

    /* enum: kObjZapBaton */
    NEW_MELEE_WEAPON(kObjZapBaton,
        "reticulan zap baton", "energy baton", "stylish baton",
        kCyan, kMeleeWeapon, kObjGenericMeleeWeapon, kPlastic,
        0, 750, kSmall, 15, 5, +1,
        newMeleeAttackP (shAttack::kZap, FULLTURN,
            100, kElectrical, 1, 6,   33, kParalyzing, 1),
        100 BZ, 10);

    /* enum: kObjMop */
    NEW_MELEE_WEAPON(kObjMop,
        "mop", "mop", "mop",
        kBrightGreen, kMeleeWeapon, kObjGenericMeleeWeapon, kWood,
        kPreidentified, 900, kMedium, 5, 10, +0,
        newMeleeAttack (shAttack::kSmash, FULLTURN, kConcussive, 1, 8),
        5 BZ, 5);

    /* enum: kObjNunchucks */
    NEW_MELEE_WEAPON(kObjNunchucks,
        "pair of nunchucks", "pair of chained sticks", "pair of chained sticks",
        kYellow, kMeleeWeapon, kObjGenericMeleeWeapon, kWood,
        0, 500, kMedium, 10, 10, +0,
        newMeleeAttack (shAttack::kClub, FULLTURN, kConcussive, 2, 9),
        50 BZ, 10);

    /* enum: kObjRazorWhip */
    NEW_MELEE_WEAPON(kObjRazorWhip,
        "razor whip", "whip", "razor whip",
        kBrightRed, kMeleeWeapon, kObjGenericMeleeWeapon, kCarbonFiber,
        kPreidentified, 1500, kMedium, 4, 10, +0,
        newMeleeAttack (shAttack::kWhip, QUICKTURN, kConcussive, 2, 5),
        250 BZ, 5);

    /* enum: kObjChainsaw */
    NEW_POWERED_MELEE_WEAPON(kObjChainsaw,
        "chainsaw", "chainsaw", "chainsaw",
        kBrightYellow, kMeleeWeapon, kObjGenericMeleeWeapon, kSteel,
        kPreidentified, 1200, kMedium, 10, 10, kObjEnergyCell, 1, -2,
        newMeleeAttack (shAttack::kCut, FULLTURN, kSlashing, 4, 14),
        500 BZ, 5);

    /* enum: kObjBo */
    NEW_MELEE_WEAPON(kObjBo, // For the ninja Turtles! By Psi!
        "bo", "stick", "wooden staff",
        kBrown, kMeleeWeapon, kObjGenericMeleeWeapon, kWood,
        0, 1000, kMedium, 5, 10, +1,
        newMeleeAttack (shAttack::kClub, FULLTURN, kConcussive, 1, 8),
        5 BZ, 5);

    /* enum: kObjCombiStick */
    NEW_GENERIC_WEAPON(kObjCombiStick,
        "combi-stick", "retractable spear", "retractable spear",
        kBrightRed, kMeleeWeapon, kMeleeWeapon, kObjGenericMeleeWeapon,
        kTitanium, 0, 2500, kMedium, 20, 10, +0,
        newMeleeAttack (shAttack::kStab, FULLTURN, kPiercing, 1, 12),
        newMissileAttack (shAttack::kStab, FULLTURN, 12, kPiercing, 4, 12),
        newAttack (shAttack::kCombi, shAttack::kExtend,
            kMelee, FULLTURN, 3, 0, kPiercing, 1, 8),
        900 BZ, 5);

    /* enum: kObjGomJabbar */
    NEW_MELEE_WEAPON(kObjGomJabbar,
        "Gom Jabbar", "needle", "poison needle",
        kWhite, kMeleeWeapon, kObjGenericMeleeWeapon, kWood,
        0, 50, kTiny, 1, 1, +0,
        newMeleeAttack (shAttack::kPrick, FULLTURN, kPoisonous, 3, 4),
        500 BZ, 1);

    /* enum: kObjKatana */
    NEW_MELEE_WEAPON(kObjKatana,
        "katana", "sword", "antique sword",
        kGray, kSword, kObjGenericMeleeWeapon, kSteel,
        0, 600, kSmall, 10, 2, +0,
        newMeleeAttack (shAttack::kSword, FULLTURN, kSlashing, 1, 10),
        1000 BZ, 10);

    /* enum: kObjChaosSword */
    NEW_MELEE_WEAPON(kObjChaosSword,
        "sword of chaos", "sword", "antique sword",
        kGray, kSword, kObjGenericMeleeWeapon, kSteel,
        0, 1000, kSmall, 10, 2, +0,
        newMeleeAttack (shAttack::kSword, FULLTURN,
            kSlashing, 4, 14, kPsychic, 1, 1),
        2000 BZ, 3);

    /* enum: kObjBatleth */
    NEW_MELEE_WEAPON(kObjBatleth,
        "bat'leth", "bat'leth", "bat'leth",
        kGray, kSword, kObjGenericMeleeWeapon, kSteel,
        0, 1000, kLarge, 20, 2, -1,
        newMeleeAttack (shAttack::kSlash, FULLTURN, kSlashing, 2, 12),
        1200 BZ, 10);

    /* enum: kObjLightSaber */
    NEW_MELEE_WEAPON(kObjLightSaber,
        "lightsaber", "energy sword", "energy sword",
        kRed, kSword, kObjGenericMeleeWeapon, kTitanium,
        0, 600, kMedium, 30, 2, +1,
        newMeleeAttack (shAttack::kSword, FULLTURN,
            kSlashing, 2, 12, kBurning, 1, 8),
        1900 BZ, 10);
    AllIlks[kObjLightSaber]->mWieldedIntrinsics |= kReflection;

    /* enum: kObjElvenSword */
    NEW_MELEE_WEAPON(kObjElvenSword, //Eldar-ish
        "elven psychic sword", "curvy blade", "organic sword",
        kMagenta, kSword, kObjGenericMeleeWeapon, kBone,
        0, 750, kSmall, 20, 2, +2,
        newMeleeAttack (shAttack::kSword, FULLTURN,
            kSlashing, 3, 5, kPsychic, 1, 1),
        1500 BZ, 10);

    /* enum: kObjGenericUnarmedWeapon */
    //Well, "unarmed" weapons!
    NEW_MELEE_WEAPON(kObjGenericUnarmedWeapon,
        "unarmed weapon", "weapon", "weapon",
        kCyan, kUnarmedCombat, kObjAnyWeapon, kSteel,
        0, 0, kTiny, 0, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjWristBlade */
    NEW_MELEE_WEAPON(kObjWristBlade,
        "wrist blade", "wrist blade", "wrist blade",
        kCyan, kUnarmedCombat, kObjGenericUnarmedWeapon, kSteel,
        kPreidentified, 500, kSmall, 15, 10, +0,
        newMeleeAttack (shAttack::kStab, FULLTURN, kPiercing, 1, 8),
        100 BZ, 5);

    /* enum: kObjPowerFist */
    NEW_MELEE_WEAPON(kObjPowerFist,
        "power fist", "mechanical glove", "armored glove",
        kGray, kUnarmedCombat, kObjGenericUnarmedWeapon, kSteel,
        0, 2000, kMedium, 20, 10, +0,
        newMeleeAttack (shAttack::kPunch, SLOWTURN,
            kConcussive, 3, 12, kStunning, 1, 3),
        1400 BZ, 5);

    /* enum: kObjPowerClaw */
    NEW_MELEE_WEAPON(kObjPowerClaw,
        "power claw", "mechanical pincers", "steel pincers",
        kGray, kUnarmedCombat, kObjGenericUnarmedWeapon, kSteel,
        0, 2500, kLarge, 15, 15, -1,
        newMeleeAttackP (shAttack::kClaw, SLOWTURN,
            100, kPiercing, 2, 14,   50, kParalyzing, 1, 2),
        350 BZ, 5);

    /* enum: kObjPhaseGloves */
    NEW_POWERED_MELEE_WEAPON(kObjPhaseGloves,
        "pair of phase gloves", "pair of unstable gloves", "pair of shimmering gloves",
        kBrightBlue, kUnarmedCombat, kObjGenericUnarmedWeapon, kCarbonFiber,
        0, 500, kTiny, 5, 5, kObjEnergyCell, 4, +2,
        newMeleeAttack (shAttack::kPunch, FULLTURN,
            kForce, 2, 8, kElectrical, 2, 5),
        2000 BZ, 5);

    /* enum: kObjPsiBlades */
    /* They drain 2 points of psionics. */
    NEW_MELEE_WEAPON(kObjPsiBlades,
        "pair of psi blades", "pair of generators", "pair of generators",
        kBrightYellow, kUnarmedCombat, kObjGenericUnarmedWeapon, kGold,
        0, 1200, kSmall, 20, 4, +2,
        newMeleeAttack (shAttack::kSlash, HALFTURN,
            kSlashing, 1, 3, kPsychic, 1, 1),
        1800 BZ, 5);

    /* enum: kObjGenericGrenade */
    NEW_MISSILE_WEAPON(kObjGenericGrenade,
        "grenade", "grenade", "grenade",
        kCyan, kObjAnyWeapon, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        NULL, 25 BZ, ABSTRACT);

    int n = 0;
    /* enum: kObjConcussionGrenade */
    NEW_MISSILE_WEAPON(kObjConcussionGrenade,
        "concussion grenade", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kBlast, 1, kConcussive, 3, 18),
        25 BZ, 100);

    ++n;
    /* enum: kObjFragGrenade */
    NEW_MISSILE_WEAPON(kObjFragGrenade,
        "frag grenade", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kBlast, 1,
            kConcussive, 2, 12, kPiercing, 2, 12),
        25 BZ, 50);

    ++n;
    /* enum: kObjStunGrenade */
    NEW_MISSILE_WEAPON(kObjStunGrenade,
        "stun grenade", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kBlast, 1, kStunning, 3, 6),
        25 BZ, 50);

    ++n;
    /* enum: kObjIncendiaryGrenade */
    NEW_MISSILE_WEAPON(kObjIncendiaryGrenade,
        "incendiary grenade", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kIncendiary, 2,
            kBurning, 3, 9, kConcussive, 2, 12),
        25 BZ, 25);

    ++n;
    /* enum: kObjRadGrenade */
    NEW_MISSILE_WEAPON(kObjRadGrenade,
        "rad grenade", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kRadioactive | kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kFlash, 3, kRadiological, 4, 32),
        25 BZ, 25);

    ++n;
    /* enum: kObjFlashbang */
    NEW_MISSILE_WEAPON(kObjFlashbang,
        "flashbang", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kLight, 5,
            kStunning, 1, 4, kBlinding, 2, 10),
        25 BZ, 50);

    ++n;
    /* enum: kObjFlare */
    NEW_MISSILE_WEAPON(kObjFlare,
        "flare", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kNoAttack, 3, kFlare, 0, 0),
        25 BZ, 25);

    ++n;
    /* enum: kObjEMPGrenade */
    NEW_MISSILE_WEAPON(kObjEMPGrenade,
        "EMP grenade", SGrenadeData[n].mDesc, GrenadeData[n].mDesc,
        kCyan, kObjGenericGrenade, kSteel,
        kMergeable, 400, kTiny, 5, 1, +0,
        newGrenadeAttack (shAttack::kBlast, 1, kMagnetic, 3, 18),
        25 BZ, 25);

    /* enum: kObjGenericAmmunition */
    NEW_AMMUNITION(kObjGenericAmmunition,
        "ammunition", "ammunition", "ammunition",
        kBrown, kObjAnyWeapon, kLead,
        kMergeable | kPreidentified, 10, kFine, 0 BZ, ABSTRACT);

    /* enum: kObjShotgunShell */
    NEW_AMMUNITION(kObjShotgunShell,
        "shotgun shell", "shotgun shell", "shotgun shell",
        kYellow, kObjGenericAmmunition, kLead,
        kMergeable | kPreidentified, 32, kFine, 1 BZ, 100);

    /* enum: kObjBullet */
    NEW_AMMUNITION(kObjBullet,
        "bullet", "bullet", "bullet",
        kYellow, kObjGenericAmmunition, kSteel,
        kMergeable | kPreidentified, 10, kFine, 1 BZ, 350);

    /* enum: kObjRailgunSlug */
    NEW_AMMUNITION(kObjRailgunSlug,
        "railgun slug", "slug", "slug",
        kRed, kObjGenericAmmunition, kDepletedUranium,
        kMergeable | kPreidentified, 80, kFine, 2 BZ, 25);

    /* enum: kObjFuelDrum */
    NEW_AMMUNITION(kObjFuelDrum,
        "fuel drum", "fuel drum", "fuel drum",
        kCyan, kObjGenericAmmunition, kSteel,
        kMergeable | kPreidentified, 10, kFine, 2 BZ, 75);

    /* enum: kObjWebCasket */
    NEW_AMMUNITION(kObjWebCasket,
        "web casket", "casket", "casket",
        kGray, kObjGenericAmmunition, kPlasteel,
        kMergeable | kPreidentified, 44, kFine, 3 BZ, 10);

    /* enum: kObjSpearheadDart */
    NEW_AMMUNITION(kObjSpearheadDart,
        "spearhead dart", "spearhead dart", "spearhead dart",
        kBrown, kObjGenericAmmunition, kPlasteel,
        kMergeable | kPreidentified, 19, kSmall, 1 BZ, 10);

    /* enum: kObjGenericGun */
    NEW_GUN(kObjGenericGun,
        "gun", "gun", "gun",
        kCyan, kNoSkillCode, kObjAnyWeapon, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjGenericConvGun */
    NEW_GUN(kObjGenericConvGun,
        "conventional gun", "gun", "gun",
        kCyan, kNoSkillCode, kObjGenericGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjGenericConvHandGun */
    NEW_GUN(kObjGenericConvHandGun,
        "handgun", "handgun", "handgun",
        kCyan, kHandgun, kObjGenericConvGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjPeaShooter */
    NEW_GUN(kObjPeaShooter,
        "pea shooter", "pistol", "wimpy pistol",
        kCyan, kHandgun, kObjGenericConvHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjNothing, 0, +0,
        newGunAttack (shAttack::kBolt, FULLTURN, 30, kForce, 1, 4),
        50 BZ, 50);

    /* enum: kObjPistol */
    NEW_GUN(kObjPistol,
        "ordinary pistol", "pistol", "ordinary pistol",
        kBlue, kHandgun, kObjGenericConvHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjBullet, 1, -1,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 8),
        100 BZ, 25);

    /* enum: kObjKhanHeavyPistol */
    NEW_GUN(kObjKhanHeavyPistol,
        "Khan heavy pistol", "pistol", "robust pistol",
        kBlue, kHandgun, kObjGenericConvHandGun, kSteel,
        0, 1050, kMedium, 10, 2, kObjBullet, 1, -2,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 3, 9),
        250 BZ, 15);

    /* enum: kObjAssaultPistol */
    NEW_GUN(kObjAssaultPistol,
        "assault pistol", "pistol", "assault pistol",
        kBlue, kHandgun, kObjGenericConvHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjBullet, 2, -1,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 8),
        100 BZ, 10);

    /* enum: kObjNetLauncher */
    NEW_GUN(kObjNetLauncher,
        "net launcher", "pistol", "wimpy pistol",
        kBlue, kHandgun, kObjGenericConvHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjWebCasket, 1, 0,
        newGunAttack (shAttack::kWeb, FULLTURN, 8, kWebbing, 1, 8),
        500 BZ, 10);

    /* enum: kObjGenericConvLightGun */
    NEW_GUN(kObjGenericConvLightGun,
        "light gun", "light gun", "light gun",
        kBlue, kLightGun, kObjGenericConvGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjShotgun */
    NEW_GUN(kObjShotgun,
        "shotgun", "shotgun", "shotgun",
        kBlue, kLightGun, kObjGenericConvLightGun, kSteel,
        kPreidentified, 1200, kLarge, 10, 3, kObjShotgunShell, 1, +1,
        newGunAttack (shAttack::kShot, FULLTURN, 10, kConcussive, 4, 24),
        150 BZ, 25);

    /* enum: kObjSawnoffShotgun */
    NEW_GUN(kObjSawnoffShotgun,
        "sawn-off shotgun", "shotgun", "sawn-off shotgun",
        kBlue, kLightGun, kObjGenericConvLightGun, kSteel,
        kPreidentified, 1050, kLarge, 10, 3, kObjShotgunShell, 1, +1,
        newConeAttack (shAttack::kShot, FULLTURN, 5, 1, kConcussive, 2, 8),
        150 BZ, 15);

    /* enum: kObjSniperRifle */
    NEW_GUN(kObjSniperRifle,
        "sniper rifle", "rifle", "telescoped rifle",
        kBlue, kLightGun, kObjGenericConvLightGun, kSteel,
        0, 1200, kMedium, 10, 2, kObjBullet, 1, +3,
        newGunAttack (shAttack::kBullet, FULLTURN, 180, kPiercing, 1, 10),
        300 BZ, 25);

    /* enum: kObjLightFlamethrower */
    NEW_GUN(kObjLightFlamethrower,
        "light flamethrower", "light flamethrower", "light flamethrower",
        kBrightCyan, kLightGun, kObjGenericConvLightGun, kSteel,
        kPreidentified, 900, kMedium, 10, 3, kObjFuelDrum, 2, +0,
        newConeAttack (shAttack::kIncendiary, FULLTURN, 6, 3, kBurning, 2, 6),
        200 BZ, 10);

    /* enum: kObjSpeargun */
    NEW_GUN(kObjSpeargun,
        "speargun rifle", "speargun rifle", "speargun rifle",
        kRed, kLightGun, kObjGenericConvLightGun, kSteel,
        kPreidentified, 1600, kMedium, 15, 5, kObjSpearheadDart, 1, +0,
        newGunAttack (shAttack::kSpear, FULLTURN, 60, kPiercing, 1, 10),
        300 BZ, 25);

    /* enum: kObjGenericConvHeavyGun */
    NEW_GUN(kObjGenericConvHeavyGun,
        "heavy gun", "heavy gun", "heavy gun",
        kCyan, kHeavyGun, kObjGenericConvGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjPulseRifle */
    NEW_GUN(kObjPulseRifle,
        "pulse rifle", "rifle", "assault rifle",
        kBlue, kHeavyGun, kObjGenericConvHeavyGun, kSteel,
        kSelectiveFire, 3000, kLarge, 10, 3, kObjBullet, 3, +0,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 8),
        400 BZ, 20);
    AllIlks[kObjPulseRifle]->mUseFunc = selectWeaponFireMode;

    /* M56A2 and M57 come with auto-aim feature.
       See Creature.cpp:shObject::applyConferredSkills for details. */
    /* enum: kObjM56Smartgun */
    NEW_GUN(kObjM56Smartgun,
        "M56 Smartgun", "rifle", "assault rifle",
        kBlue, kHeavyGun, kObjGenericConvHeavyGun, kSteel,
        kSelectiveFire, 3500, kLarge, 10, 3, kObjBullet, 3, +1,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 9),
        700 BZ, 10);
    AllIlks[kObjM56Smartgun]->mUseFunc = selectWeaponFireMode;

    /* enum: kObjM57Smartgun */
    NEW_GUN(kObjM57Smartgun,
        "M57 Smartgun", "rifle", "assault rifle",
        kBlue, kHeavyGun, kObjGenericConvHeavyGun, kSteel,
        kSelectiveFire, 3500, kLarge, 10, 3, kObjBullet, 3, +1,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 9),
        700 BZ, 10);
    AllIlks[kObjM57Smartgun]->mUseFunc = selectWeaponFireMode;

/*
    NEW_GUN(kObjBoltgun,
        "boltgun", "pistol", "assault pistol",
        kCyan, kHeavyGun, NULL, kSteel,
        0, 1000, kMedium, 10, 3, kObjBullet, 3, +0,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 10),
        500 BZ, 25);
*/
    /* enum: kObjRailgun */
    NEW_GUN(kObjRailgun,
        "railgun", "railgun", "railgun",
        kRed, kHeavyGun, kObjGenericConvHeavyGun, kSteel,
        kPreidentified, 2000, kMedium, 10, 3, kObjRailgunSlug, 1, +0,
        newGunAttack (shAttack::kRail, SLOWTURN, 180, kPiercing, 3, 30),
        800 BZ, 15);

    /* enum: kObjChaingun */
    NEW_GUN(kObjChaingun,
        "chaingun", "heavy cannon", "barreled cannon",
        kYellow, kHeavyGun, kObjGenericConvHeavyGun, kSteel,
        kPreidentified, 3000, kLarge, 15, 5, kObjBullet, 6, +0,
        newGunAttack (shAttack::kBullet, FULLTURN, 60, kPiercing, 1, 10),
        800 BZ, 10);

    /* enum: kObjFlamethrower */
    NEW_GUN(kObjFlamethrower,
        "flamethrower", "flamethrower", "flamethrower",
        kBrightCyan, kHeavyGun, kObjGenericConvHeavyGun, kSteel,
        kPreidentified, 1400, kLarge, 10, 3, kObjFuelDrum, 5, +0,
        newConeAttack (shAttack::kIncendiary, FULLTURN, 8, 2, kBurning, 3, 6),
        300 BZ, 10);

    /* enum: kObjGenericEnergyGun */
    NEW_GUN(kObjGenericEnergyGun,
        "energy gun", "gun", "gun",
        kCyan, kNoSkillCode, kObjGenericGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjGenericEnergyHandGun */
    NEW_GUN(kObjGenericEnergyHandGun,
        "energy handgun", "handgun", "handgun",
        kCyan, kHandgun, kObjGenericEnergyGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjLaserPistol */
    NEW_GUN(kObjLaserPistol,
        "laser pistol", "pistol", "energy pistol",
        kCyan, kHandgun, kObjGenericEnergyHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjEnergyCell, 5, -2,
        newBeamAttack (shAttack::kLaser, FULLTURN, 30,
            kLaser, 1, 7, kBlinding, 1, 10),
        50 BZ, 25);

    /* enum: kObjBlaster */
    NEW_GUN(kObjBlaster,
        "blaster pistol", "pistol", "energy pistol",
        kCyan, kHandgun, kObjGenericEnergyHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjEnergyCell, 5, -1,
        newGunAttack (shAttack::kBolt, FULLTURN, 30, kForce, 1, 8),
        50 BZ, 25);

    /* enum: kObjPhaser */
    NEW_GUN(kObjPhaser,
        "phaser pistol", "pistol", "ergonomic energy pistol",
        kCyan, kHandgun, kObjGenericEnergyHandGun, kSteel,
        0, 800, kMedium, 10, 2, kObjEnergyCell, 5, +1,
        newBeamAttack (shAttack::kLaser, FULLTURN, 30,
            kLaser, 1, 8, kBlinding, 1, 10),
        50 BZ, 15);

/*
        PLASMA WEAPONS:
        I've set them to have a burning and magnetic effect since plasma
        is superheated ionized gas, I hope that doesn't make them too
        overpowered against droids. --Psi
*/
    /* enum: kObjPlasmaPistol */
    NEW_GUN(kObjPlasmaPistol,
        "plasma pistol", "pistol", "coiled pistol",
        kGreen, kHandgun, kObjGenericEnergyHandGun, kSteel,
        0, 900, kMedium, 15, 5, kObjEnergyCell, 6, +0,
        newGunAttack (shAttack::kPlasmaGlob, FULLTURN, 30,
            kBurning, 1, 10, kMagnetic, 1, 6),
        150 BZ, 15);

    /* enum: kObjZapGun */
    NEW_GUN(kObjZapGun,
        "reticulan zap gun", "pistol", "stylized energy pistol",
        kGray, kHandgun, kObjGenericEnergyHandGun, kPlastic,
        0, 800, kMedium, 10, 2, kObjEnergyCell, 4, +1,
        newGunAttackP (shAttack::kBolt, FULLTURN, 30,
            100, kElectrical, 1, 6,   40, kParalyzing, 1, 2),
        50 BZ, 10);

    /* enum: kObjPlasmaCaster */
    NEW_GUN(kObjPlasmaCaster,
        "plasma caster", "pistol", "coiled pistol",
        kGreen, kHandgun, kObjGenericEnergyHandGun, kSteel,
        0, 1200, kMedium, 20, 5, kObjEnergyCell, 9, -8,
        newGunAttack (shAttack::kPlasmaGlob, FULLTURN, 30,
            kBurning, 1, 12, kMagnetic, 1, 6),
        1275 BZ, 5);
    AllIlks[kObjPlasmaCaster]->mFlags |= kChargeable;
    AllIlks[kObjPlasmaCaster]->mMaxCharges = 2;
    AllIlks[kObjPlasmaCaster]->mPsiModifier = -1;
    AllIlks[kObjPlasmaCaster]->mUseFunc = chargePlasmaCaster;
    AllIlks[kObjPlasmaCaster]->mZapAttack =
        newAttack (shAttack::kBlast, shAttack::kFarBurst, kAimed,
            SLOWTURN, 30, 1, kBurning, 3, 15, kMagnetic, 2, 10),
    /* ZapAttack radius may be tuned by shootWeapon in Fight.cpp. */

    /* enum: kObjGenericEnergyLightGun */
    NEW_GUN(kObjGenericEnergyLightGun,
        "energy light gun", "light gun", "light gun",
        kCyan, kLightGun, kObjGenericEnergyGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjLaserRifle */
    NEW_GUN(kObjLaserRifle,
        "laser rifle", "rifle", "energy rifle",
        kCyan, kLightGun, kObjGenericEnergyLightGun, kSteel,
        0, 1200, kLarge, 10, 3, kObjEnergyCell, 8, +1,
        newBeamAttack (shAttack::kLaser, FULLTURN, 60,
            kLaser, 1, 11, kBlinding, 3, 15),
        200 BZ, 75);

    /* enum: kObjPlasmaRifle */
    NEW_GUN(kObjPlasmaRifle,
        "plasma rifle", "rifle", "coiled rifle",
        kGreen, kLightGun, kObjGenericEnergyLightGun, kSteel,
        0, 1600, kMedium, 15, 5, kObjEnergyCell, 10, +1,
        newGunAttack (shAttack::kPlasmaGlob, FULLTURN, 45,
            kBurning, 2, 16, kMagnetic, 1, 6),
        450 BZ, 10);

    /* enum: kObjGenericEnergyHeavyGun */
    NEW_GUN(kObjGenericEnergyHeavyGun,
        "energy heavy gun", "heavy gun", "heavy gun",
        kCyan, kHeavyGun, kObjGenericEnergyGun, kSteel,
        0, 100, kTiny, 10, 1, kObjNothing, 0, +0,
        NULL, 0 BZ, ABSTRACT);

    /* enum: kObjMiningLaser */
    NEW_GUN(kObjMiningLaser,
        "mining laser", "heavy cannon", "energy cannon",
        kCyan, kHeavyGun, kObjGenericEnergyHeavyGun, kSteel,
        0, 3750, kLarge, 15, 5, kObjEnergyCell, 25, +8,
        newBeamAttackP (shAttack::kLaser, SLOWTURN, 0 /* +d6 */,
            100, kLaser, 2, 22,   80, kBlinding, 1, 4,   50, kDigging, 0, 0),
        800 BZ, 20);

    /* enum: kObjLaserCannon */
    NEW_GUN(kObjLaserCannon,
        "laser cannon", "heavy cannon", "energy cannon",
        kCyan, kHeavyGun, kObjGenericEnergyHeavyGun, kSteel,
        0, 3000, kLarge, 15, 5, kObjEnergyCell, 35, +1,
        newBeamAttack (shAttack::kLaser, SLOWTURN, 60,
            kLaser, 4, 44, kBlinding, 1, 4),
        800 BZ, 10);

    /* enum: kObjPlasmaCannon */
    /* enum: kObjLastWeapon = kObjPlasmaCannon */
    //I obviously based this cannon on Cyrus' one, but tweaked
    //it in order to keep it in style with the other plasma weapons. --Psi
    NEW_GUN(kObjPlasmaCannon,
        "plasma cannon", "heavy cannon", "coiled cannon",
        kGreen, kHeavyGun, kObjGenericEnergyHeavyGun, kSteel,
        0, 5000, kLarge, 15, 5, kObjEnergyCell, 40, +0,
        newGunAttack (shAttack::kPlasmaGlob, SLOWTURN, 60,
            kBurning, 4, 40, kMagnetic, 8, 20),
        1200 BZ, 5);
}


int
shObject::isAmmo (shObject *weapon)
{
    if (NULL == weapon) {
        return isA (kProjectile);
    }
    return isA (weapon->myIlk ()->mAmmoType);
}


/* returns: 1 if the creature has enough ammo to shoot the weapon once,
              or if the weapon doesn't need ammo; 0 o/w
 */
int
shCreature::hasAmmo (shObject *weapon)
{
    shObjectIlk *ilk = weapon->myIlk ();
    shObjId ammo = weapon->isA (kRayGun) ? kObjNothing : ilk->mAmmoType;
    int i;
    int n = 0;

    if (kObjNothing == ammo) {
        if (weapon->isChargeable ()) {
            if (weapon->mCharges) {
                return 1;
            } else {
                if (isHero ()) {
                    weapon->setChargeKnown ();
                }
                return 0;
            }
        }
        return 1;
    }
    n = ilk->mAmmoBurst;
    if (kObjEnergyCell == ammo) {
        return countEnergy () >= n ? 1 : 0;
    }
    for (i = 0; n > 0 && i < mInventory->count (); i++) {
        shObject *obj = mInventory->get (i);
        if (obj->isA (ammo)) {
            n -= obj->mCount;
        }
    }
    return n <= 0 ? 1 : 0;
}


//RETURNS: number of rounds taken on success; 0 if no ammo available;
//         -1 if none were needed
//         For weapons that use energy cells, returns 1 or 0 depending on
//          whether or not enough cells were consumed

int
shCreature::expendAmmo (shObject *weapon, int cnt /* = 0 */ )
{
    shObjectIlk *ilk = weapon->myIlk ();
    shObjId ammo = ilk->mAmmoType;
    int i;
    int n = 0;

    if (0 == cnt) {
        if (weapon->isSelectiveFireWeapon () && !weapon->isToggled ()) {
            cnt = 1;
        } else {
            cnt = ilk->mAmmoBurst;
        }
    }
    if (weapon->isChargeable () && !weapon->isA (kObjPlasmaCaster)) {
        if (weapon->mCharges > cnt) {
            weapon->mCharges -= cnt;
        } else {
            cnt = weapon->mCharges;
            weapon->mCharges = 0;
        }
        return cnt;
    } else if (kObjNothing == ammo) {
        return -1;
    }
    if (kObjEnergyCell == ammo) {
        if (countEnergy () >= cnt) {
            loseEnergy (cnt);
            return 1;
        } else {
            return 0;
        }
    }

    for (i = 0; n < cnt && i < mInventory->count (); i++) {
        shObject *obj = mInventory->get (i);
        if (obj->isA (ammo)) {
            if (kObjBullet != ammo && obj->mCount < cnt) {
                return 0; /* Only bullet weapons may burst right now. */
            }
            int x = useUpSomeObjectsFromInventory (obj, cnt - n);
            if (x) {
                n += x;
                --i;
            }
        }
    }
    return n;
}


static int
selectWeaponFireMode (shObject *gun)
{
    if (gun->isToggled ()) {
        I->p ("You switch your weapon to single fire mode.");
        gun->resetToggled ();
    } else {
        I->p ("You switch your weapon to burst mode.");
        gun->setToggled ();
    }
    return HALFTURN;
}

static int /* Mostly a YAFM. */
useAnalProbe (shObject *probe)
{
    if (Hero.mGlyph.mSym == 'g') { /* Reticulan. */
        Hero.sufferDamage (probe->myIlk ()->mMeleeAttack);
        probe->identify ();
        I->p ("Your experience tells you it is %s.", probe->inv ());
        return FULLTURN;
    } else {
        I->p ("You sick bastard! I am not going to do it!");
        return 0;
    }
}

static int
chargePlasmaCaster (shObject *caster)
{
    const int CHARGE_COST[2] = {24, 36}; /* Energy cells needed. */
    caster->setChargeKnown ();
    if (caster->mCharges == 2) {
        I->p ("You cannot charge it further.");
        return 0;
    } else {
        int next = caster->mCharges;
        if (Hero.countEnergy () >= CHARGE_COST[next]) {
            Hero.loseEnergy (CHARGE_COST[next]);
            ++caster->mCharges;
            caster->setChargeKnown ();
            I->p ("You charge the plasma caster.");
            return caster->mCharges * FULLTURN;
        } else {
            I->p ("You need %d energy to add %s charge.", CHARGE_COST[next],
                  next ? "second" : "first");
            return 0;
        }
    }
}
